SHOW_BROWSER = False
SIGNED_COOKIES = False


def pytest_addoption(parser):
    parser.addoption("--show-browser", action="store_true", default=False, help="Show web browser window")
    parser.addoption("--signed-cookies", action="store_true", default=False, help="Use signed cookies session backend")


def pytest_configure(config):
    global SHOW_BROWSER, SIGNED_COOKIES
    SHOW_BROWSER = config.option.show_browser
    SIGNED_COOKIES = config.option.signed_cookies



import uuid

from django import forms
from django.http import HttpResponseRedirect
from django.shortcuts import render
from django.utils.html import mark_safe

from .models import Thing

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


def test_misc(request):
    return render(
        request,
        "tests/test_misc.html",
        {"name": request.session.get("name", None)},
    )


def redirect_to_misc(request):
    return HttpResponseRedirect(reverse("test_misc"))


def set_sess_foo_to_bar(request):
    request.session["foo"] = "bar"
    return render(request, "tests/test_misc.html")


class AddSpacersMixin:
    def __init__(self, add_spacers=False, **kwargs):
        super().__init__(**kwargs)
        self.add_spacers = add_spacers

    def as_p(self):
        retval = super().as_p()
        if self.add_spacers:
            # Hack to help test interacting with elements
            # that aren't in view.
            retval = mark_safe(retval.replace("</p>", "</p>" + ("<br>" * 100)))
        return retval


class ThingForm(AddSpacersMixin, forms.ModelForm):
    category = forms.ChoiceField(choices=Thing.CATEGORY_CHOICES, widget=forms.RadioSelect)

    class Meta:
        model = Thing
        fields = [
            "name",
            "big",
            "clever",
            "element_type",
            "category",
            "count",
            "description",
        ]


class ThingFormWithSelectForCategory(AddSpacersMixin, forms.ModelForm):
    class Meta:
        model = Thing
        fields = ThingForm._meta.fields


# Have separate forms so that we test different form enctype
class ThingFormWithUpload(AddSpacersMixin, forms.ModelForm):
    class Meta:
        model = Thing
        fields = ["name", "notes_file"]


def edit_thing(request, thing_id, with_upload=False):
    thing = Thing.objects.get(id=int(thing_id))
    add_spacers = "add_spacers" in request.GET
    add_js_delay = int(request.GET.get("add_js_delay", "0"))

    if with_upload:
        form_class = ThingFormWithUpload
        redirect_url = reverse("edit_thing_with_upload", kwargs={"thing_id": thing_id})
    else:
        select_for_category = "select_for_category" in request.GET
        form_class = ThingFormWithSelectForCategory if select_for_category else ThingForm
        redirect_url = reverse("edit_thing", kwargs={"thing_id": thing_id})

    if request.method == "POST":
        if "clear" in request.POST:
            thing = Thing(id=thing.id, category=Thing.CATEGORY_MAGMA)
            thing.save()
            return HttpResponseRedirect(reverse("thing_cleared", kwargs={"thing_id": thing_id}))
        else:
            thing_form = form_class(
                data=request.POST,
                files=request.FILES,
                instance=thing,
                add_spacers=add_spacers,
            )
            if thing_form.is_valid():
                thing_form.save()
                return HttpResponseRedirect(redirect_url)
    else:
        thing_form = form_class(instance=thing, add_spacers=add_spacers)

    return render(
        request,
        "tests/edit_thing.html",
        {
            "thing_form": thing_form,
            "thing": thing,
            "add_js_delay": add_js_delay,
            "upload": with_upload,
        },
    )


def edit_thing_with_upload(request, thing_id):
    return edit_thing(request, thing_id, with_upload=True)


def auto_submit_form(request):
    return render(
        request,
        "tests/auto_submit_form.html",
        {
            "method": request.method,
            "chosen": request.POST["food"] if request.method == "POST" else None,
        },
    )


def delayed_appearance(request):
    return render(
        request,
        "tests/delayed_appearance.html",
        {
            "add_js_delay": request.GET.get("add_js_delay", 0),
        },
    )


def list_things(request):
    things = Thing.objects.all()
    if "select_thing" in request.GET:
        id_list = map(int, request.GET.getlist("select_thing"))
        selected_things = things.filter(id__in=id_list)
    else:
        selected_things = []
    return render(
        request,
        "tests/list_things.html",
        {
            "things": things,
            "selected_things": selected_things,
        },
    )


def thing_cleared(request, thing_id):
    thing = Thing.objects.get(id=int(thing_id))
    return render(request, "tests/thing_cleared.html", {"thing": thing})


def new_browser_session_test(request):
    if "UID" in request.session:
        uid = request.session["UID"]
        message = "Welcome back"
    else:
        uid = uuid.uuid1()
        request.session["UID"] = str(uid)
        message = "Hello new user"

    return render(
        request,
        "tests/new_browser_session_test.html",
        {
            "uid": uid,
            "message": message,
        },
    )


def unscrollable(request):
    count = int(request.POST.get("count", 0))
    name = request.POST.get("name", "")
    item = request.POST.get("itemdropdown", "")
    return render(
        request,
        "tests/unscrollable.html",
        {
            "count": count,
            "nextcount": count + 1,
            "name": name,
            "item": item,
        },
    )



import os.path

import conftest

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": "main.db",
        "CONN_MAX_AGE": 0,
        "TEST": {
            "NAME": "tests.db",
        },
    }
}


DEBUG = True
USE_TZ = True
ROOT_URLCONF = "django_functest_tests.urls"
INSTALLED_APPS = [
    # First, see http://stackoverflow.com/questions/18281137/selenium-django-gives-foreign-key-error/18292090#18292090 # noqa
    "django.contrib.contenttypes",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django_functest",
    "django_functest_tests",
]
SITE_ID = 1
MIDDLEWARE = [
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]
STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(os.path.abspath(os.path.dirname(__file__)), "..", "..", "/tmp_static/")
TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": True,
    "root": {
        "level": "WARNING",
        "handlers": ["console"],
    },
    "formatters": {
        "verbose": {"format": "%(levelname)s %(asctime)s %(module)s %(process)d %(thread)d %(message)s"},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        },
    },
}
SECRET_KEY = "foo"


if conftest.SIGNED_COOKIES:
    SESSION_ENGINE = "django.contrib.sessions.backends.signed_cookies"



import warnings
from importlib import import_module

from django.conf import settings
from django.contrib.auth import BACKEND_SESSION_KEY as AUTH_BACKEND_SESSION_KEY
from django.contrib.auth import HASH_SESSION_KEY as AUTH_HASH_SESSION_KEY
from django.contrib.auth import SESSION_KEY as AUTH_ID_SESSION_KEY
from django.contrib.auth import authenticate
from django.utils.html import escape
from furl import furl
from lxml import etree


class _NotPassed:
    pass


NotPassed = _NotPassed()


class ShortcutLoginMixin:
    """
    A mixin that provides a fast way of logging in and out.
    """

    def shortcut_login(self, user=None, **credentials):
        """
        Login a user directly. Pass the user object, without a password,
        or pass in credentials like: username="me", password="my_password"
        """
        if user is not None:
            if credentials:
                raise AssertionError("Either pass 'user' or **credentials, not both")

            def get_backend():
                from django.contrib.auth import load_backend

                for backend_path in settings.AUTHENTICATION_BACKENDS:
                    backend = load_backend(backend_path)
                    if hasattr(backend, "get_user"):
                        return backend_path

            user.backend = get_backend()
        else:
            user = authenticate(**credentials)
            if not user:
                raise ValueError(f"User {user} was not authenticated")

        session_auth_hash = ""
        if hasattr(user, "get_session_auth_hash"):
            session_auth_hash = user.get_session_auth_hash()

        # Mimicking django.contrib.auth functionality
        self.set_session_data(
            {
                AUTH_ID_SESSION_KEY: user.pk,
                AUTH_HASH_SESSION_KEY: session_auth_hash,
                AUTH_BACKEND_SESSION_KEY: user.backend,
            }
        )

    def shortcut_logout(self):
        self.flush_session()


def get_session_store(session_key=None):
    engine = import_module(settings.SESSION_ENGINE)
    # Implement a database session store object that will contain the session key.
    store = engine.SessionStore(session_key=session_key)
    if session_key is None:
        store.save()
    else:
        store.load()
    return store


class CommonMixin:
    def assertUrlsEqual(self, url, other_url=None):
        """
        Asserts that the URLs match. Empty protocol or domain are ignored.
        """
        if other_url is None:
            other_url = self.current_url
        url1 = furl(url)
        url2 = furl(other_url)
        self.assertEqual(url1.path, url2.path)
        self.assertEqual(url1.query, url2.query)
        if url1.netloc and url2.netloc:
            self.assertEqual(url1.netloc, url2.netloc)
        if url1.scheme and url2.scheme:
            self.assertEqual(url1.scheme, url2.scheme)

    def fill_by_id(self, data):
        """
        Same as ``fill`` except the keys are input IDs
        """
        warnings.warn(
            "instead of `fill_by_id({'foo': 'bar'})` do `fill({'#foo': 'bar'})`",
            DeprecationWarning,
        )
        self.fill({"#" + k: v for k, v in data.items()})

    def fill_by_name(self, fields, prefix="", scroll=NotPassed):
        """
        Same as ``fill`` except the keys are input names
        """
        self.fill({f'[name="{prefix}{k}"]': v for k, v in fields.items()}, scroll=scroll)

    def _assertTextPresent(self, text, pyquery_obj, within):
        norm_text = html_norm(escape(text))
        matching_elements = pyquery_obj.find(within)
        if len(matching_elements) == 0:
            self.fail(f"No elements matched the CSS selector {within!r}")
        elif len(matching_elements) == 1:
            # Better error message for the common case:
            self.assertIn(
                norm_text,
                etree.tostring(matching_elements[0], encoding="unicode"),
            )
        else:
            self.assertTrue(
                any(norm_text in etree.tostring(elem, encoding="unicode") for elem in matching_elements),
                "Didn't find {!r} inside any of the {} matching elements for {!r}".format(
                    text, len(matching_elements), within
                ),
            )

    def _assertTextAbsent(self, text, pyquery_obj, within):
        norm_text = html_norm(escape(text))
        matching_elements = pyquery_obj.find(within)
        if len(matching_elements) == 1:
            # Better error message for the common case:
            self.assertNotIn(
                norm_text,
                etree.tostring(matching_elements[0], encoding="unicode"),
            )
        else:
            self.assertFalse(
                any(norm_text in etree.tostring(elem, encoding="unicode") for elem in matching_elements),
                "Didn't find {!r} inside any of the {} matching_elements for {!r}".format(
                    text,
                    len(matching_elements),
                    within,
                ),
            )


class AdminLoginMixin(ShortcutLoginMixin):
    """
    A mixin that logs in via the normal admin login page
    """

    def do_login(self, username=None, password=None, shortcut=True):
        if shortcut:
            self.shortcut_login(username=username, password=password)
            return

        self.get_url("admin:index")
        self.fill(
            {
                "#id_username": username,
                "#id_password": password,
            }
        )
        self.submit("input[type=submit]")

    def do_logout(self, shortcut=True):
        if shortcut:
            self.shortcut_logout()
            return

        self.get_url("admin:logout")


class BrowserSessionToken:
    # Simple container class to hide the fact that the value is really a
    # webdriver instance or webtest instance (but that may change in future)
    def __init__(self, value):
        self.value = value

    def __eq__(self, other):
        return self.value == other.value

    def __repr__(self):
        return f"<BrowserSessionToken {id(self.value)}>"


def html_norm(html):
    return html.replace("&quot;", '"').replace("&apos;", "'").replace("&#39;", "'").replace("&#x27;", "'")



from django_functest import FuncBaseMixin, ShortcutLoginMixin

from .base import SeleniumTestBase, WebTestBase
from .factories import create_user


class UserAdminBase(FuncBaseMixin, ShortcutLoginMixin):
    def test_change_self_details(self):
        user = create_user()  # noqa
        ...  # TODO


class UserAdminWT(UserAdminBase, WebTestBase):
    pass


class UserAdminSL(UserAdminBase, SeleniumTestBase):
    pass



import django

try:
    from django.urls import path

    url = None
except ImportError:
    path = None
    from django.conf.urls import url

from django.http import HttpResponse


def emptypage(request):
    # Minimal page needed for some tests.

    # We include a favicon link to stop browsers making a request to
    # /favicon.ico which slows things down (potentially a lot) and can cause
    # other issues.
    return HttpResponse(
        """<html>
    <head>
        <link href="data:image/gif;base64,R0lGODlhEAAQAIABAACE/////yH5BAEKAAEALA"""
        """AAAAAQABAAAAIghI9pwe2+nmRxvmobzmFnb4GTKJEXwEFoSq2sqSqyUQAAOw==" rel="shortcut icon">
    </head>
    <body></body>
</html>"""
    )


if path is None:
    urlpatterns = [
        url(r"^__emptypage/$", emptypage, name="django_functest.emptypage"),
    ]
else:
    urlpatterns = [
        path(r"__emptypage/", emptypage, name="django_functest.emptypage"),
    ]


if django.VERSION < (1, 9):
    from django.conf.urls import patterns

    urlpatterns = patterns("", *urlpatterns)



import os

import pytest
from django.contrib.auth import get_user_model
from selenium.common.exceptions import NoSuchElementException

from django_functest import AdminLoginMixin, FuncBaseMixin

from .base import ChromeBase, FirefoxBase
from .models import Thing

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


# Tests for Selenium specific methods


class FuncSeleniumSpecificBase(AdminLoginMixin, FuncBaseMixin):
    def setUp(self):
        super().setUp()
        self.thing = Thing.objects.create(
            name="Rock",
            big=True,
            clever=False,
            element_type=Thing.ELEMENT_EARTH,
            category=Thing.CATEGORY_MAGMA,
        )
        User = get_user_model()
        self.user = User.objects.create_superuser("admin", "admin@example.com", "password")

    def test_is_element_displayed(self):
        self.get_url("admin:login")
        assert self.is_element_displayed("#id_username")
        assert not self.is_element_displayed("#id_something_else")
        self.execute_script("document.querySelector('#id_username').style.display = 'none';")
        assert not self.is_element_displayed("#id_username")

    def test_click(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.click("button[name=check]")
        self.assertTextPresent("Everything is fine")

    def test_click_xpath(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.click(xpath='//button[@name="check"]')
        self.assertTextPresent("Everything is fine")

    def test_click_text(self):
        self.get_url("list_things")
        self.click(text="Edit Rock", text_parent_id="id_thinglist", wait_for_reload=True)
        self.assertUrlsEqual(reverse("edit_thing", kwargs={"thing_id": self.thing.id}))

    def test_click_text_missing(self):
        self.get_url("list_things")
        with pytest.raises(NoSuchElementException):
            self.click(text="Edit Fridge", wait_timeout=0)
        with pytest.raises(NoSuchElementException):
            self.click(text="Edit Rock", text_parent_id="fribble", wait_timeout=0)

    def test_double_click(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.click("button[name=check]", double=True)
        self.assertTextPresent("Everything is really fine")

    def test_double_click_element_that_changes(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.click("[name=debounced]", double=True)
        self.assertTextPresent("Pressed x 1")

    def test_double_click_element_that_disappears(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.click("[name=disappears]", double=True)
        self.assertTextPresent("Pressed x 1")

    def test_execute_script(self):
        self.get_url("test_misc")
        assert self.execute_script("return 1 + 1;") == 2

    def test_execute_script_with_args(self):
        self.get_url("test_misc")
        retval = self.execute_script("return arguments[0] + arguments[1];", 1, 2)
        assert retval == 3

    def test_hover(self):
        self.get_url("test_misc")
        get_style = "return document.defaultView.getComputedStyle(document.querySelector('#hoverable'))['font-style']"
        assert self.execute_script(get_style) == "normal"
        self.hover("#hoverable")
        assert self.execute_script(get_style) == "italic"

    def test_switch_window(self):
        self.do_login(username="admin", password="password")
        self.get_url("admin:auth_user_change", self.user.id)
        self.click("#add_id_groups")
        old_window, new_window = self.switch_window()
        self.fill({"#id_name": "My new group"})
        self.switch_window(old_window)
        self.fill({"#id_first_name": "My first name"})
        self.switch_window(new_window)
        self.submit("input[name=_save]", window_closes=True)
        self.switch_window(old_window)
        self.submit("input[name=_save]")

        User = get_user_model()
        user = User.objects.get(id=self.user.id)
        assert [g.name for g in user.groups.all()] == ["My new group"]
        assert user.first_name == "My first name"

    def test_save_screenshot(self):
        testname = f"tests.test_selenium.{self.__class__.__name__}.test_save_screenshot"

        fname = None
        try:
            fname = self.save_screenshot()
            assert testname in fname
            assert os.path.exists(fname)
        finally:
            if fname is not None:
                os.unlink(fname)


class TestFuncSeleniumSpecificFirefox(FuncSeleniumSpecificBase, FirefoxBase):
    pass


class TestFuncSeleniumSpecificChrome(FuncSeleniumSpecificBase, ChromeBase):
    pass


# Test class attribute `browser_window_size` works correctly:


class BrowserSizeBase:
    browser_window_size = (800, 700)

    def _get_window_size(self):
        if self._driver.name == "phantomjs":
            return self.execute_script("return [document.width, document.height]")
        else:
            return self.execute_script("return [window.outerWidth, window.outerHeight]")

    def test_size(self):
        width, height = self._get_window_size()
        assert 795 < width < 805
        assert 695 < height < 705

    def test_resize(self):
        self.set_window_size(700, 500)
        width, height = self._get_window_size()
        assert 695 < width < 705
        assert 495 < height < 505


class TestBrowserSizeFirefox(BrowserSizeBase, FirefoxBase):
    pass


class TestBrowserSizeChrome(BrowserSizeBase, ChromeBase):
    pass



from webtest import Upload as WebTestUpload


class Upload(WebTestUpload):
    """
    A file to upload::

        >>> Upload('filename.txt', b'data')
        <Upload "filename.txt">
    """

    # We deliberately override the docstring of WebTestUpload above,
    # to provide help that works for Selenium too.



import urllib
from collections import defaultdict

from django.conf import settings
from django_webtest import WebTestMixin
from pyquery.pyquery import PyQuery
from webtest.forms import Checkbox

from .base import FuncBaseMixin
from .exceptions import WebTestCantUseElement, WebTestMultipleElementsException, WebTestNoSuchElementException
from .utils import BrowserSessionToken, CommonMixin, NotPassed, get_session_store

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


class FuncWebTestMixin(WebTestMixin, CommonMixin, FuncBaseMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._all_last_responses = defaultdict(list)
        self._all_apps = []

    # Public Common API
    def assertTextAbsent(self, text, within="body"):
        """
        Asserts that the text is not present within the body of the current page,
        or within any element matching the CSS selector passed as `within`.
        """
        self._assertTextAbsent(text, self._make_pq(self.last_response), within)

    def assertTextPresent(self, text, within="body", wait=True):
        """
        Asserts that the text is present within the body of the current page,
        or within an element matching the CSS selector passed as `within`.
        """
        self._assertTextPresent(text, self._make_pq(self.last_response), within)

    def back(self):
        """
        Go back in the browser.
        """
        self.last_responses.pop()

    @property
    def current_url(self):
        """
        The current full URL
        """
        return self.last_response.request.url

    def follow_link(self, css_selector):
        """
        Follows the link specified in the CSS selector.
        """
        elems = self._make_pq(self.last_response).find(css_selector)
        if len(elems) == 0:
            raise WebTestNoSuchElementException(f"Can't find element matching '{css_selector}'")

        hrefs = []
        for e in elems:
            if "href" in e.attrib:
                hrefs.append(e.attrib["href"])

        if not hrefs:
            raise WebTestCantUseElement(f"No href attribute found for '{css_selector}'")

        if not all(h == hrefs[0] for h in hrefs):
            raise WebTestMultipleElementsException(
                f"Different href values for links '{css_selector}': '{' ,'.join(hrefs)}'"
            )
        final_url = urllib.parse.urljoin(self.current_url, hrefs[0])
        self.get_literal_url(final_url)

    def fill(self, data, scroll=NotPassed):
        """
        Fills form inputs using the values in fields, which is a dictionary
        of CSS selectors to values.
        """
        for selector, value in data.items():
            form, field_name, elem = self._find_form_and_field_by_css_selector(self.last_response, selector)
            field_items = form.fields[field_name]
            if isinstance(field_items, list) and len(field_items) > 1:
                # We've got something like a set of checkboxes with the same name.
                selected_value = elem.attrib["value"]
                for checkbox in field_items:
                    if checkbox._value == selected_value:
                        checkbox.checked = value
            else:
                form[field_name] = value

    def fill_by_text(self, fields, scroll=NotPassed):
        """
        Same as ``fill`` except the values are text captions. Useful for ``select`` elements.
        """
        for selector, text in fields.items():
            form, field_name, _ = self._find_form_and_field_by_css_selector(self.last_response, selector)
            self._fill_field_by_text(form, field_name, text)

    def get_element_attribute(self, css_selector, attribute):
        """
        Returns the value of the attribute of the element matching the css_selector,
        or None if there is no such element or attribute.
        """
        elems = self._make_pq(self.last_response).find(css_selector)
        if len(elems) == 0:
            return None
        if len(elems) > 1:
            raise WebTestMultipleElementsException(f"Multiple elements found matching '{css_selector}'")
        return elems[0].attrib.get(attribute, None)

    def get_element_inner_text(self, css_selector):
        """
        Returns the "inner text" (innerText in JS) of the element matching
        the css_selector, or None if there is none.
        """
        elems = self._make_pq(self.last_response).find(css_selector)
        if len(elems) == 0:
            return None
        if len(elems) > 1:
            raise WebTestMultipleElementsException(f"Multiple elements found matching '{css_selector}'")
        return inner_text(elems[0])

    def get_url(self, name, *args, **kwargs):
        """
        Gets the named URL, passing *args and **kwargs to Django's URL 'reverse' function.
        """
        return self.get_literal_url(reverse(name, args=args, kwargs=kwargs))

    def get_literal_url(self, url, auto_follow=True, expect_errors=False):
        """
        Gets the passed in URL, as a literal relative URL, without using reverse.
        """
        return self._get_url_raw(url, auto_follow=auto_follow, expect_errors=expect_errors)

    def is_element_present(self, css_selector):
        """
        Returns True if the element specified by the CSS selector is present on the current page,
        False otherwise.
        """
        return len(self._make_pq(self.last_response).find(css_selector)) > 0

    @property
    def is_full_browser_test(self):
        """
        True for Selenium tests, False for WebTest tests.
        """
        return False

    def get_session_data(self):
        """
        Returns the current Django session dictionary
        """
        return dict(self._get_session())

    def set_session_data(self, item_dict):
        """
        Set a dictionary of items directly into the Django session.
        """
        session = self._get_session()
        for name, value in item_dict.items():
            session[name] = str(value)
        session.save()
        self._update_session_cookie(session)  # Required for signed_cookie backend

    def new_browser_session(self):
        """
        Creates (and switches to) a new session that is separate from previous
        sessions. Returns a tuple (old_session_token, new_session_token). These
        values should be treated as opaque tokens that can be used with
        switch_browser_session.
        """
        # WebTestMixin creates the instance as 'self.app', so we just just move
        # that value around.
        last_app = self.app
        self.renew_app()
        return (BrowserSessionToken(last_app), BrowserSessionToken(self.app))

    def switch_browser_session(self, session_token):
        """
        Switch to the browser session indicated by the supplied token.
        Returns a tuple (old_session_token, new_session_token).
        """
        last_app = self.app
        self.app = session_token.value
        return (BrowserSessionToken(last_app), BrowserSessionToken(self.app))

    def submit(self, css_selector, wait_for_reload=None, auto_follow=True, window_closes=None, scroll=NotPassed):
        """
        Submit the form. css_selector should refer to a form, or a button/input to use
        to submit the form.
        """
        try:
            form = self._find_form_by_css_selector(self.last_response, css_selector)
            field_name = None
        except WebTestNoSuchElementException:
            form, field_name, _ = self._find_form_and_field_by_css_selector(
                self.last_response,
                css_selector,
                require_name=False,
                filter_selector="input[type=submit], button",
            )

        response = form.submit(field_name)
        if auto_follow:
            while 300 <= response.status_int < 400:
                response = response.follow()
        self.last_responses.append(response)

    def value(self, css_selector):
        """
        Returns the value of the form input specified in the CSS selector
        """
        form, field_name, _ = self._find_form_and_field_by_css_selector(
            self.last_response, css_selector, require_name=False
        )
        field = form[field_name]
        if isinstance(field, Checkbox):
            return field.checked
        else:
            return field.value

    # WebTest specific

    @property
    def last_response(self):
        """
        Returns the last WebTest response received.
        """
        return self.last_responses[-1]

    # Semi-public (used by mixins)

    def flush_session(self):
        session = self._get_session()
        session.flush()
        self._update_session_cookie(session)  # Required for signed_cookie backend

    # Implementation methods - private
    @property
    def last_responses(self):
        return self._all_last_responses[self.app]

    def _delete_cookie(self, name):
        # Awkward cookiejar interface
        for cookie in self.app.cookiejar:
            if cookie.name == name:
                self.app.cookiejar.clear(name=cookie.name, path=cookie.path, domain=cookie.domain)

    def _set_cookie(self, name, value):
        self.app.set_cookie(name, value)

    def _get_session(self):
        session_key = self.app.cookies.get(settings.SESSION_COOKIE_NAME, None)
        if session_key is None:
            # Create new
            session = get_session_store()
            self._update_session_cookie(session)
        else:
            session_key = session_key.strip('"')
            session = get_session_store(session_key=session_key)
        return session

    def _update_session_cookie(self, session):
        if session.is_empty():
            self._delete_cookie(settings.SESSION_COOKIE_NAME)
        else:
            self._set_cookie(settings.SESSION_COOKIE_NAME, session.session_key)

    def _get_url_raw(self, url, auto_follow=True, expect_errors=False):
        """
        'raw' method for getting URL - not compatible between FullBrowserTest and WebTestBase
        """
        self.last_responses.append(self.app.get(url, auto_follow=auto_follow, expect_errors=expect_errors))
        return self.last_response

    def _find_form_by_css_selector(self, response, css_selector):
        pq = self._make_pq(response)
        items = pq.find(css_selector)
        if any(item.tag == "form" for item in items):
            if len(items) > 1:
                raise WebTestMultipleElementsException(f"Found multiple forms matching {css_selector}")
            return self._match_form_elem_to_webtest_form(items[0], response)
        else:
            raise WebTestNoSuchElementException(f"Can't find form matching {css_selector}")

    def _find_form_and_field_by_css_selector(self, response, css_selector, filter_selector=None, require_name=True):
        pq = self._make_pq(response)
        items = pq.find(css_selector)

        found = []
        if filter_selector:
            items = items.filter(filter_selector)
        for item in items:
            form_elem = self._find_parent_form(item)
            if form_elem is None:
                raise WebTestCantUseElement(f"Can't find form for input {css_selector}.")
            form = self._match_form_elem_to_webtest_form(form_elem, response)
            field = item.name if hasattr(item, "name") else item.attrib.get("name", None)
            if field is None and require_name:
                raise WebTestCantUseElement(f"Element {css_selector} needs 'name' attribute in order to use it")
            found.append((form, field, item))

        if len(found) > 1:
            if not all(f[0:2] == found[0][0:2] for f in found):
                raise WebTestMultipleElementsException(f"Multiple elements found matching '{css_selector}'")

        if len(found) > 0:
            return found[0]

        raise WebTestNoSuchElementException(f"Can't find element matching {css_selector} in response {response}.")

    def _find_parent_form(self, elem):
        p = elem.getparent()
        if p is None:
            return None
        if p.tag == "form":
            return p
        return self._find_parent_form(p)

    def _fill_field_by_text(self, form, field_name, text):
        field = form[field_name]
        if field.tag == "select":
            for val, _, t in field.options:
                if t == text:
                    form[field_name] = val
                    break
            else:
                raise ValueError(f"No option matched '{text}'")
        else:
            raise WebTestCantUseElement(f"Don't know how to 'fill_by_text' for elements of type '{field.tag}'")

    def _match_form_elem_to_webtest_form(self, form_elem, response):
        pq = self._make_pq(response)
        forms = pq("form")
        form_index = forms.index(form_elem)
        webtest_form = response.forms[form_index]
        form_sig = {
            "action": form_elem.attrib.get("action", ""),
            "id": form_elem.attrib.get("id", ""),
            "method": form_elem.attrib.get("method", "").lower(),
        }
        webtest_sig = {
            "action": getattr(webtest_form, "action", ""),
            "id": getattr(webtest_form, "id", ""),
            "method": getattr(webtest_form, "method", "").lower(),
        }
        webtest_sig = {k: v if v is not None else "" for k, v in webtest_sig.items()}
        assert form_sig == webtest_sig
        return webtest_form

    def _make_pq(self, response):
        # Cache to save parsing every time
        if not hasattr(self, "_pq_cache"):
            self._pq_cache = {}
        if response in self._pq_cache:
            return self._pq_cache[response]
        # Don't use `response.pyquery` because of https://github.com/Pylons/webtest/issues/245
        pq = PyQuery(response.testbody, parser="html" if "html" in response.content_type else "xml")
        self._pq_cache[response] = pq
        return pq


def inner_text(elem, root=True):
    return (elem.text or "") + "".join(inner_text(e, root=False) for e in elem) + ("" if root else (elem.tail or ""))



import subprocess
import unittest

import conftest
import pytest
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test import TestCase, override_settings

from django_functest import FuncSeleniumMixin, FuncWebTestMixin, MultiThreadedLiveServerMixin

# Getting some errors that seem related to this:
# http://stackoverflow.com/questions/18281137/selenium-django-gives-foreign-key-error/18292090#18292090 # noqa
# producing a stacktrace in which a postmigrate handler attempts to
# create permissions as part of fixture teardown and fails.
# This is not fixed by change the order of INSTALLED_APPS. However,
# we can avoid the code path by using TestCase.available_apps

AVAILABLE_APPS = [
    "django.contrib.contenttypes",
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.staticfiles",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django_functest",
    "django_functest_tests",
]


def binary_available(filename):
    return subprocess.call(["which", filename], stdout=subprocess.PIPE) == 0


firefox_available = binary_available("firefox")
chrome_available = binary_available("chromedriver")


@override_settings(DEBUG=True)  # easier debugging
class MyLiveServerTestCase(StaticLiveServerTestCase):
    available_apps = AVAILABLE_APPS


class WebTestBase(FuncWebTestMixin, TestCase):
    available_apps = AVAILABLE_APPS


class HideBrowserMixin:
    display = conftest.SHOW_BROWSER


@pytest.mark.selenium
class SeleniumBaseMixin:
    browser_window_size = (1024, 768)


@unittest.skipIf(not firefox_available, "Firefox not available, skipping")
@pytest.mark.firefox
class FirefoxBase(HideBrowserMixin, SeleniumBaseMixin, FuncSeleniumMixin, MyLiveServerTestCase):
    driver_name = "Firefox"

    firefox_binary = None  # default, hacked by runtests.py

    @classmethod
    def get_webdriver_options(cls):
        kwargs = super().get_webdriver_options()
        if cls.firefox_binary is not None:
            from selenium.webdriver.firefox.firefox_binary import FirefoxBinary

            kwargs["firefox_binary"] = FirefoxBinary(firefox_path=cls.firefox_binary)
        return kwargs


@unittest.skipIf(not chrome_available, "Chrome not available, skipping")
@pytest.mark.chrome
class ChromeBase(
    HideBrowserMixin,
    SeleniumBaseMixin,
    FuncSeleniumMixin,
    MultiThreadedLiveServerMixin,
    MyLiveServerTestCase,
):
    driver_name = "Chrome"



import django
from django.contrib import admin

try:
    from django.urls import include, path

    url = None
except ImportError:
    path = None
    from django.conf.urls import url, include

from . import views

if path is None:
    urlpatterns = [
        url(r"^admin/", admin.site.urls),
        url(r"^django_functest/", include("django_functest.urls")),
        url(r"^test_misc/", views.test_misc, name="test_misc"),
        url(
            r"^redirect_to_misc/",
            views.redirect_to_misc,
            name="redirect_to_misc",
        ),
        url(
            r"^set_sess_foo_to_bar/",
            views.set_sess_foo_to_bar,
            name="set_sess_foo_to_bar",
        ),
        url(r"^list_things/", views.list_things, name="list_things"),
        url(r"^edit_thing/(?P<thing_id>.*)/", views.edit_thing, name="edit_thing"),
        url(
            r"^edit_thing_with_upload/(?P<thing_id>.*)/",
            views.edit_thing_with_upload,
            name="edit_thing_with_upload",
        ),
        url(
            r"^thing_cleared/(?P<thing_id>.*)/",
            views.thing_cleared,
            name="thing_cleared",
        ),
        url(
            r"^new_browser_session_test/",
            views.new_browser_session_test,
            name="new_browser_session_test",
        ),
        url(
            r"^auto_submit_form/",
            views.auto_submit_form,
            name="auto_submit_form",
        ),
        url(
            r"^delayed_appearance/",
            views.delayed_appearance,
            name="delayed_appearance",
        ),
        url(
            r"^unscrollable/",
            views.unscrollable,
            name="unscrollable",
        ),
    ]
else:
    urlpatterns = [
        path("admin/", admin.site.urls),
        path("django_functest/", include("django_functest.urls")),
        path("test_misc/", views.test_misc, name="test_misc"),
        path(
            "redirect_to_misc/",
            views.redirect_to_misc,
            name="redirect_to_misc",
        ),
        path(
            "set_sess_foo_to_bar/",
            views.set_sess_foo_to_bar,
            name="set_sess_foo_to_bar",
        ),
        path("list_things/", views.list_things, name="list_things"),
        path("edit_thing/<str:thing_id>/", views.edit_thing, name="edit_thing"),
        path(
            "edit_thing_with_upload/<str:thing_id>/",
            views.edit_thing_with_upload,
            name="edit_thing_with_upload",
        ),
        path("thing_cleared/<str:thing_id>/", views.thing_cleared, name="thing_cleared"),
        path(
            "new_browser_session_test/",
            views.new_browser_session_test,
            name="new_browser_session_test",
        ),
        path(
            r"auto_submit_form/",
            views.auto_submit_form,
            name="auto_submit_form",
        ),
        path(
            r"delayed_appearance/",
            views.delayed_appearance,
            name="delayed_appearance",
        ),
        path(
            r"unscrollable/",
            views.unscrollable,
            name="unscrollable",
        ),
    ]

if django.VERSION < (1, 9):
    from django.conf.urls import patterns

    urlpatterns = patterns("", *urlpatterns)



import itertools
from typing import Any

from ..models import User

# Factory utils:


class _Auto:
    """
    Sentinel value (used instead of None because that can be a valid value)
    """

    def __bool__(self):
        return False

    def __repr__(self):
        return "<Auto>"


Auto: Any = _Auto()


def str_sequence(template: str, start_at: int = 0):
    """
    Makes a sequence of strings using an integer substution, counting from start_at.
    """
    return (template.format(idx) for idx in itertools.count())


# User factory:

USERNAME_SEQUENCE = str_sequence("_auto_user_{0}")


def create_user(username: str = Auto, is_superuser: bool = Auto, is_staff: bool = Auto):
    username = username or next(USERNAME_SEQUENCE)
    if is_superuser is Auto:
        is_superuser = False
    if is_staff is Auto:
        is_staff = is_superuser

    return User.objects.create(username=username, is_superuser=is_superuser, is_staff=is_staff)



#!/usr/bin/env python

import os
import os.path
import sys

import django
from django.conf import settings
from django.core.management import execute_from_command_line

settings_dict = dict(
    DATABASES={
        "default": {
            "ENGINE": "django.db.backends.sqlite3",
            "NAME": "main.db",
            "CONN_MAX_AGE": 0,
            "TEST": {
                "NAME": "tests.db",
            },
        }
    },
    INSTALLED_APPS=[
        "django_functest",
        "django_functest_tests",
    ],
)


settings.configure(**settings_dict)


try:
    setup = django.setup
except AttributeError:
    pass
else:
    setup()

initial_migration = "tests/django_functest_tests/migrations/0001_initial.py"
if os.path.exists(initial_migration):
    os.unlink(initial_migration)
argv = [sys.argv[0], "makemigrations", "tests"] + sys.argv[2:]
execute_from_command_line(argv)



BROWSER = "Firefox"
SHOW_BROWSER = False


def pytest_addoption(parser):
    parser.addoption(
        "--browser", type=str, default="Firefox", help="Selenium driver_name to use", choices=["Firefox", "Chrome"]
    )
    parser.addoption("--show-browser", action="store_true", default=False, help="Show web browser window")


def pytest_configure(config):
    global SHOW_BROWSER, BROWSER
    BROWSER = config.option.browser
    SHOW_BROWSER = config.option.show_browser



from .base import FuncBaseMixin
from .files import Upload
from .funcselenium import FuncSeleniumMixin
from .funcwebtest import FuncWebTestMixin
from .server import MultiThreadedLiveServerMixin
from .utils import AdminLoginMixin, ShortcutLoginMixin

__version__ = "1.5.1"

__all__ = [
    "FuncBaseMixin",
    "FuncWebTestMixin",
    "FuncSeleniumMixin",
    "ShortcutLoginMixin",
    "AdminLoginMixin",
    "MultiThreadedLiveServerMixin",
    "Upload",
]


FuncCommonApi = FuncBaseMixin



# Generated by Django 4.1 on 2022-08-22 13:43

import django.contrib.auth.models
import django.contrib.auth.validators
import django.utils.timezone
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                ("last_login", models.DateTimeField(blank=True, null=True, verbose_name="last login")),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={"unique": "A user with that username already exists."},
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[django.contrib.auth.validators.UnicodeUsernameValidator()],
                        verbose_name="username",
                    ),
                ),
                ("first_name", models.CharField(blank=True, max_length=150, verbose_name="first name")),
                ("last_name", models.CharField(blank=True, max_length=150, verbose_name="last name")),
                ("email", models.EmailField(blank=True, max_length=254, verbose_name="email address")),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. "
                        "Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                ("date_joined", models.DateTimeField(default=django.utils.timezone.now, verbose_name="date joined")),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to"
                        " each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "abstract": False,
            },
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
    ]



#!/usr/bin/env python
import os
import sys

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = "1.5.1"

if sys.argv[-1] == "publish":
    try:
        import wheel  # noqa
    except ImportError:
        print('Wheel library missing. Please run "pip install wheel"')
        sys.exit()
    os.system("python setup.py sdist upload")
    os.system("python setup.py bdist_wheel upload")
    sys.exit()

if sys.argv[-1] == "tag":
    print("Tagging the version on github:")
    os.system(f"git tag -a {version} -m 'version {version}'")
    os.system("git push --tags")
    sys.exit()

readme = open("README.rst").read()

setup(
    name="django-functest",
    version=version,
    description="""Helpers for creating functional tests in Django, with a unified API for WebTest and Selenium tests.""",  # noqa
    long_description=readme,
    author="Luke Plant",
    author_email="L.Plant.98@cantab.net",
    url="https://github.com/django-functest/django-functest",
    packages=[
        "django_functest",
    ],
    package_dir={
        "django_functest": "src/django_functest",
    },
    python_requires=">=3.7",
    include_package_data=True,
    install_requires=[
        "django-webtest>=1.9.9",
        "WebTest>=3",
        "selenium>=4",
        "furl>=0.4.9",
        "pyquery>=1.2.10",
        "Django>=2.0",
    ],
    license="BSD",
    zip_safe=False,
    keywords="django-functest",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Framework :: Django",
        "Framework :: Django :: 2.0",
        "Framework :: Django :: 2.1",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.0",
        "Framework :: Django :: 3.1",
        "Framework :: Django :: 3.2",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
)



# Generated by Django 4.0.1 on 2022-08-19 15:49

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Thing",
            fields=[
                ("id", models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255)),
                ("big", models.BooleanField(default=False)),
                ("clever", models.BooleanField(default=False)),
                (
                    "element_type",
                    models.CharField(
                        choices=[("e", "Earth"), ("w", "Water"), ("a", "Air"), ("f", "Fire")], max_length=1
                    ),
                ),
                ("category", models.IntegerField(choices=[(0, "Magma"), (1, "Quasigroup"), (2, "Monoid")], default=0)),
                ("count", models.IntegerField(default=0)),
                ("description", models.TextField(blank=True)),
                ("notes_file", models.FileField(blank=True, upload_to="django_functest/tests/uploads")),
            ],
        ),
    ]



import pytest
from selenium.common.exceptions import NoSuchElementException, TimeoutException

from django_functest import FuncBaseMixin, Upload
from django_functest.exceptions import (
    SeleniumCantUseElement,
    WebTestCantUseElement,
    WebTestMultipleElementsException,
    WebTestNoSuchElementException,
)

from .base import ChromeBase, FirefoxBase, WebTestBase
from .models import Thing

try:
    from django.urls import reverse
except ImportError:
    from django.core.urlresolvers import reverse


class CommonBase(FuncBaseMixin):
    def setUp(self):
        super().setUp()
        self.thing = Thing.objects.create(
            name="Rock",
            big=True,
            clever=False,
            element_type=Thing.ELEMENT_EARTH,
            category=Thing.CATEGORY_MAGMA,
            count=1,
            description="Hard thing",
        )

    def test_get_url(self):
        self.get_url("admin:login")
        url = self.current_url
        assert url.endswith("/admin/login/")
        assert url.startswith("http://")

    def test_get_literal_url(self):
        url = reverse("admin:login")
        self.get_literal_url(url)
        self.assertUrlsEqual(url)

    def test_get_literal_url_with_full_url(self):
        url = reverse("admin:login")
        self.get_literal_url(url)
        # Specifically check this idiom for refreshing a page:
        self.get_literal_url(self.current_url)
        self.assertUrlsEqual(url)

    def test_assertUrlsEqual_default(self):
        self.get_url("admin:login")
        with pytest.raises(AssertionError):
            self.assertUrlsEqual("foo")
        self.assertUrlsEqual("/admin/login/")

    def test_assertUrlsEqual_path(self):
        with pytest.raises(AssertionError):
            self.assertUrlsEqual("/login/", "/admin/login/")
        self.assertUrlsEqual("/login/", "/login/")

    def test_assertUrlsEqual_query(self):
        with pytest.raises(AssertionError):
            self.assertUrlsEqual("/foo/?q=1", "/foo/")
        with pytest.raises(AssertionError):
            self.assertUrlsEqual("/foo/?q=1", "/foo/?q=2")
        self.assertUrlsEqual("/foo/?q=1", "/foo/?q=1")

    def test_assertUrlsEqual_host(self):
        self.assertUrlsEqual("/foo/", "//example.com/foo/")
        self.assertUrlsEqual("//example.com/foo/", "//example.com/foo/")
        with pytest.raises(AssertionError):
            self.assertUrlsEqual("//example.com/foo/", "//other.com/foo/")

    def test_assertUrlsEqual_protocol(self):
        self.assertUrlsEqual("http://example.com/foo/", "//example.com/foo/")
        self.assertUrlsEqual("http://example.com/foo/", "http://example.com/foo/")
        with pytest.raises(AssertionError):
            self.assertUrlsEqual("http://example.com/foo/", "https://example.com/foo/")

    def test_assertTextPresent(self):
        self.get_url("test_misc")
        self.assertTextPresent("Hello world")
        # Check escaping
        # NB we are using " here, not the &quot; found in the template source.
        self.assertTextPresent("from 'me' & \"friends\"")
        self.assertTextPresent("""It's also allowed to have "quotes" without escaping in text in valid HTML""")

        with pytest.raises(AssertionError):
            self.assertTextPresent("Something definitely not there")

    def test_assertTextAbsent(self):
        self.get_url("test_misc")
        self.assertTextAbsent("Something definitely not there")
        with pytest.raises(AssertionError):
            self.assertTextAbsent("Hello world")
        with pytest.raises(AssertionError):
            self.assertTextAbsent("from 'me' & \"friends\"")
        with pytest.raises(AssertionError):
            self.assertTextAbsent(
                """It's also allowed to have "quotes" """ """without escaping in text in valid HTML"""
            )

    def test_assertTextPresent_within(self):
        self.get_url("test_misc")
        self.assertTextPresent("Hello world", within="p")
        with pytest.raises(AssertionError):
            self.assertTextPresent("Hello world", within="p.myclass")
        with pytest.raises(AssertionError, match="No elements matched the CSS selector 'p.not-a-real-class'"):
            self.assertTextPresent("Hello world", within="p.not-a-real-class", wait=False)

    def test_assertTextAbsent_within(self):
        self.get_url("test_misc")
        self.assertTextAbsent("Hello world", within="p.myclass")
        with pytest.raises(AssertionError):
            self.assertTextAbsent("Hello world", within="p")

        self.assertTextAbsent("Hello world", within="p#this-is-not-a-dom-node")

    def test_assertTextAbsent_script(self):
        # This is a test that we are parsing script tags properly.
        self.get_literal_url(reverse("delayed_appearance") + "?add_js_delay=100")
        self.assertTextAbsent("Hello!", within="#new_stuff")

    def test_current_url(self):
        self.get_url("admin:login")
        # Check it really is a full URL
        assert self.current_url.startswith("http")

    def test_is_element_present(self):
        self.get_url("admin:login")
        assert self.is_element_present("#id_username")
        assert not self.is_element_present("#id_something_not_there")

    def refresh_thing(self):
        self.thing = Thing.objects.get(id=self.thing.id)
        return self.thing

    def test_fill(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.fill(
            {
                "#id_name": "New name",
                "#id_big": False,
                "#id_clever": True,
                "#id_element_type": Thing.ELEMENT_AIR,
                "#id_category_1": Thing.CATEGORY_QUASIGROUP,
                "#id_count": 5,
                "#id_description": "Soft thing\r\nwith line breaks",
            }
        )
        self.submit("input[name=change]")
        self._assertThingChanged()

    def test_fill_by_id(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.fill_by_id(
            {
                "id_name": "New name",
                "id_big": False,
                "id_clever": True,
                "id_element_type": Thing.ELEMENT_AIR,
                "id_category_1": Thing.CATEGORY_QUASIGROUP,
                "id_count": 5,
                "id_description": "Soft thing\r\nwith line breaks",
            }
        )
        self.submit("input[name=change]")
        self._assertThingChanged()

    def test_fill_by_name(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.fill_by_name(
            {
                "name": "New name",
                "big": False,
                "clever": True,
                "element_type": Thing.ELEMENT_AIR,
                "category": Thing.CATEGORY_QUASIGROUP,
                "count": 5,
                "description": "Soft thing\r\nwith line breaks",
            }
        )
        self.submit("input[name=change]")
        self._assertThingChanged()

    def test_fill_by_text(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.fill_by_text({"#id_element_type": "Water"})
        self.submit("input[name=change]")
        self.refresh_thing()
        assert self.thing.element_type == Thing.ELEMENT_WATER

    def test_fill_by_text_missing(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        with pytest.raises(self.TextNotFoundException):
            self.fill_by_text({"#id_element_type": "Plasma"})

    def test_fill_by_text_for_unsupported(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        with pytest.raises(self.ElementUnusableException):
            self.fill_by_text({"#id_count": "Water"})

    def _assertThingChanged(self):
        thing = self.refresh_thing()
        assert thing.name == "New name"
        assert not thing.big
        assert thing.clever
        assert thing.element_type == Thing.ELEMENT_AIR
        assert thing.category == Thing.CATEGORY_QUASIGROUP
        assert thing.count == 5
        assert thing.description == "Soft thing\r\nwith line breaks"

    def test_fill_no_element_error(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        with pytest.raises(self.ElementNotFoundException):
            self.fill({"#id_blahblah": "New name"})

    def test_fill_select_by_integer(self):
        url = reverse("edit_thing", kwargs=dict(thing_id=self.thing.id)) + "?select_for_category=1"
        self.get_literal_url(url)
        self.fill_by_name(
            {
                "name": "New name",
                "big": False,
                "clever": True,
                "element_type": Thing.ELEMENT_AIR,
                "category": Thing.CATEGORY_QUASIGROUP,
                "count": 5,
                "description": "Soft thing\r\nwith line breaks",
            }
        )
        self.submit("input[name=change]")
        self._assertThingChanged()

    def test_fill_checkbox_with_same_name(self):
        # It is allowed to have two checkboxes with the same name. It should result
        # in a submission with repeated values of
        self.thing2 = Thing.objects.create(
            name="Flower",
            big=False,
            clever=True,
            element_type=Thing.ELEMENT_EARTH,
            category=Thing.CATEGORY_MONOID,
            count=1,
            description="Soft thing",
        )
        self.thing3 = Thing.objects.create(
            name="Water",
            big=False,
            clever=False,
            element_type=Thing.ELEMENT_WATER,
            category=Thing.CATEGORY_MAGMA,
            count=1,
            description="Wet thing",
        )
        self.get_url("list_things")

        # Select:
        self.fill(
            {
                f'input[name=select_thing][value="{self.thing.id}"]': True,
                f'input[name=select_thing][value="{self.thing2.id}"]': True,
            }
        )
        self.submit("input[name=select]")
        self.assertTextPresent(f"{self.thing.name} is selected")
        self.assertTextPresent(f"{self.thing2.name} is selected")
        self.assertTextAbsent(f"{self.thing3.name} is selected")

        # Unselect:
        self.fill({f'input[name=select_thing][value="{self.thing.id}"]': False})
        self.submit("input[name=select]")
        self.assertTextAbsent(f"{self.thing.name} is selected")
        self.assertTextPresent(f"{self.thing2.name} is selected")
        self.assertTextAbsent(f"{self.thing3.name} is selected")

    def test_fill_no_scroll(self):
        self.get_literal_url(reverse("unscrollable"))

        # scroll makes no different to WebTest, but we should
        # still be able to specify it for compatibility

        # With scroll=False
        self.fill({"[name=name]": "Peter"}, scroll=False)
        self.fill_by_text({"[name=itemdropdown]": "Item 2"}, scroll=False)
        self.submit("[type=submit]")
        self.assertTextPresent("Name submitted: Peter")
        self.assertTextPresent("Item submitted: item-2")

        # With auto_scroll_by_default=False
        self.auto_scroll_by_default = False
        self.fill({"[name=name]": "Annabelle"})
        self.fill_by_text({"[name=itemdropdown]": "Item 2"})
        self.submit("[type=submit]")
        self.assertTextPresent("Name submitted: Annabelle")
        self.assertTextPresent("Item submitted: item-2")

    def test_submit(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.submit("button[name=clear]")
        thing = self.refresh_thing()
        assert thing.name == ""

    def test_submit_no_scroll(self):
        self.get_literal_url(reverse("unscrollable"))

        # scroll makes no different to WebTest, but we should
        # still be able to specify it for compatibility

        # With scroll=False to submit
        self.submit("[type=submit]", scroll=False)
        self.assertTextPresent("Submitted so far: 1")

        # With auto_scroll_by_default=False
        self.auto_scroll_by_default = False
        self.submit("[type=submit]")
        self.assertTextPresent("Submitted so far: 2")

    def test_submit_form(self):
        self.get_url("auto_submit_form")
        self.assertTextPresent("Method used: GET", within="#method")
        self.fill({"select": "ice-cream"})
        if not self.is_full_browser_test:
            # Need to manually submit
            self.submit("#the-form")
        self.assertTextPresent("Method used: POST", within="#method")
        self.assertTextPresent("You chose ice-cream")

        # Test that we can use the same API with Selenium
        self.submit("#the-form")

    def test_follow_link(self):
        self.get_url("list_things")
        self.follow_link("a.edit")
        self.assertUrlsEqual(reverse("edit_thing", kwargs={"thing_id": self.thing.id}))

    def test_follow_link_not_found(self):
        self.get_url("list_things")
        with pytest.raises(self.ElementNotFoundException):
            self.follow_link("a.foobar")

    def test_follow_link_path_relative(self):
        self.get_url("test_misc")
        self.follow_link('a[href="."]')
        self.assertUrlsEqual(reverse("test_misc"))

        self.follow_link('a[href="?param1=val1"]')
        self.assertUrlsEqual(reverse("test_misc") + "?param1=val1")

        self.follow_link('a[href="?param1=val2&param2"]')
        self.assertUrlsEqual(reverse("test_misc") + "?param1=val2&param2")

        self.follow_link('a[href="."]')  # This should clear queries
        self.assertUrlsEqual(reverse("test_misc"))

    def test_back(self):
        self.get_url("list_things")
        self.follow_link("a.edit")
        self.back()
        self.assertUrlsEqual(reverse("list_things"))

    def test_multiple_back(self):
        # We could test the behaviour regarding forms, especially those that
        # submit to the same URL and then redirect to the same URL. However,
        # Firefox and Chrome behave differently here - Firefox produces
        # fewer history entries.
        self.get_url("list_things")
        self.follow_link("a.edit")
        edit_url = reverse("edit_thing", kwargs={"thing_id": self.thing.id})
        self.assertUrlsEqual(edit_url)
        self.submit("button[name=clear]")
        self.assertUrlsEqual(reverse("thing_cleared", kwargs={"thing_id": self.thing.id}))
        self.assertTextPresent("was cleared")
        self.back()
        self.assertUrlsEqual(edit_url)
        self.back()
        self.assertUrlsEqual(reverse("list_things"))

    def test_set_session_data(self):
        self.set_session_data({"name": "The Jabberwocky"})
        self.get_url("test_misc")
        self.assertTextPresent("Hello to The Jabberwocky")

    def test_get_session_data(self):
        self.get_url("set_sess_foo_to_bar")
        sess_dict = self.get_session_data()
        assert sess_dict == {"foo": "bar"}

    def test_value(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        assert self.value("#id_name") == "Rock"
        assert self.value("#id_big") is True
        assert self.value("#id_clever") is False
        assert self.value("#id_element_type") == "e"
        assert self.value("[name=category]") == str(Thing.CATEGORY_MAGMA)
        assert self.value("#id_description") == "Hard thing"

    def test_value_immediately_after_fill(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.fill_by_name(
            {
                "name": "Some changed name",
                "big": False,
                "clever": True,
                "element_type": "w",
                "category": Thing.CATEGORY_MONOID,
                "description": "Some changed description",
            }
        )
        assert self.value("#id_name") == "Some changed name"
        assert self.value("#id_big") is False
        assert self.value("#id_clever") is True
        assert self.value("#id_element_type") == "w"
        assert self.value("#id_description") == "Some changed description"
        assert self.value("[name=category]") == str(Thing.CATEGORY_MONOID)

    def test_file_upload(self):
        self.get_url("edit_thing_with_upload", thing_id=self.thing.id)
        data = b"This is my data"
        self.fill({"#id_notes_file": Upload("notes.txt", content=data)})
        self.submit("[name=change]")
        thing = self.refresh_thing()
        assert thing.notes_file.file.read() == data

    def test_new_browser_session(self):
        self.get_url("new_browser_session_test")
        self.assertTextPresent("Hello new user")
        self.assertTextAbsent("Welcome back")
        uid_1 = self.get_session_data()["UID"]

        # Sanity check our view behaves as expected
        self.get_url("new_browser_session_test")
        self.assertTextPresent("Welcome back")
        uid_1b = self.get_session_data()["UID"]

        assert uid_1 == uid_1b

        first_session_token, second_session_token = self.new_browser_session()
        self.get_url("new_browser_session_test")
        self.assertTextPresent("Hello new user")
        self.assertTextAbsent("Welcome back")
        uid_2 = self.get_session_data()["UID"]
        assert uid_1 != uid_2

        # Tests for switch_browser_session
        ot2, nt2 = self.switch_browser_session(first_session_token)
        assert nt2 == first_session_token
        assert ot2 == second_session_token

        self.get_url("new_browser_session_test")
        self.assertTextPresent("Welcome back")
        self.assertTextPresent(uid_1)
        assert self.get_session_data()["UID"] == uid_1

        self.switch_browser_session(second_session_token)
        self.get_url("new_browser_session_test")
        self.assertTextPresent("Welcome back")
        self.assertTextPresent(uid_2)
        assert self.get_session_data()["UID"] == uid_2

        # assertTextPresent (etc.) should work without refetching
        # a page. This requires things like `last_response`
        # automatically switching.

        self.switch_browser_session(first_session_token)
        self.assertTextPresent(uid_1)
        self.assertTextAbsent(uid_2)
        self.switch_browser_session(second_session_token)
        self.assertTextPresent(uid_2)
        self.assertTextAbsent(uid_1)

    def test_get_element_inner_text(self):
        self.get_url("test_misc")
        assert self.get_element_inner_text("#inner-text-test-1") == "A paragraph with “this” & that 😄"
        assert self.get_element_inner_text("#inner-text-test-2") == "Some text with bold and italic."
        assert self.get_element_inner_text("#inner-text-test-3") == ""
        assert self.get_element_inner_text("#does-not-exist-3") is None

    def test_get_element_attribute(self):
        self.get_url("test_misc")
        assert self.get_element_attribute("#inner-text-test-1", "id") == "inner-text-test-1"
        # href tests are important here, because of the difference between
        # Selenium's `WebElement.get_attribute` and `get_dom_attribute`
        assert self.get_element_attribute("#self-link-3", "href") == "?param1=val2&param2"
        assert self.get_element_attribute("#self-link-3", "does-not-exist") is None
        assert self.get_element_attribute("#does-not-exist-3", "id") is None


class TestFuncWebTestCommon(CommonBase, WebTestBase):

    ElementNotFoundException = WebTestNoSuchElementException
    TextNotFoundException = ValueError
    ElementUnusableException = WebTestCantUseElement

    def test_is_full_browser_attribute(self):
        assert self.is_full_browser_test is False

    def test_fill_multiple_matches(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        with pytest.raises(WebTestMultipleElementsException):
            self.fill({"input[type=checkbox]": True})

    def test_fill_element_without_name(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        with pytest.raises(WebTestCantUseElement):
            self.fill({"#id_badinput1": "Hello"})

    def test_fill_element_outside_form(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        with pytest.raises(WebTestCantUseElement):
            self.fill({"#id_badinput2": "Hello"})

    def test_submit_no_auto_follow(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.submit("input[name=change]", auto_follow=False)
        assert self.last_response.status_int == 302

    def test_follow_link_multiple_matches(self):
        Thing.objects.create(name="Another")
        self.get_url("list_things")
        with pytest.raises(WebTestMultipleElementsException):
            self.follow_link("a.edit")

    def test_follow_link_no_href(self):
        self.get_url("list_things")
        with pytest.raises(WebTestCantUseElement):
            self.follow_link("a.javascriptonly")

    def test_get_literal_url_auto_follow(self):
        url = "/redirect_to_misc/"
        self.get_literal_url(url, auto_follow=True)
        self.assertUrlsEqual("/test_misc/")

        self.get_literal_url(url, auto_follow=False)
        self.assertUrlsEqual(url)
        assert self.last_response.status_int == 302

    def test_get_literal_url_expect_errors(self):
        url = "/a_404_url/"
        self.get_literal_url(url, expect_errors=True)
        assert self.last_response.status_int == 404
        with pytest.raises(Exception):
            self.get_literal_url(url, expect_errors=False)

    def test_get_element_inner_text_multiple(self):
        self.get_url("test_misc")
        with pytest.raises(WebTestMultipleElementsException):
            self.get_element_inner_text("p")

    def test_get_element_attribute_multiple(self):
        self.get_url("test_misc")
        with pytest.raises(WebTestMultipleElementsException):
            self.get_element_attribute("a", "id")

    def test_submit_form_multiple_matching(self):
        self.get_url("auto_submit_form")
        with pytest.raises(WebTestMultipleElementsException):
            self.submit("form")


class FuncSeleniumCommonBase(CommonBase):

    ElementNotFoundException = TimeoutException
    TextNotFoundException = NoSuchElementException
    ElementUnusableException = SeleniumCantUseElement

    def test_is_full_browser_attribute(self):
        assert self.is_full_browser_test is True

    def test_fill_with_scrolling(self):
        url = reverse("edit_thing", kwargs=dict(thing_id=self.thing.id)) + "?add_spacers=1"
        self.get_literal_url(url)
        self.fill(
            {
                "#id_name": "New name",
                "#id_big": False,
                "#id_clever": True,
                "#id_element_type": Thing.ELEMENT_AIR,
                "#id_category_1": Thing.CATEGORY_QUASIGROUP,
                "#id_count": 5,
                "#id_description": "Soft thing\r\nwith line breaks",
            }
        )
        self.submit("input[name=change]")
        self._assertThingChanged()

    def test_submit_no_wait_for_reload(self):
        self.get_url("edit_thing", thing_id=self.thing.id)
        self.submit("button[name=check]", wait_for_reload=False)
        self.assertTextPresent("Everything is fine")

    def test_submit_slow_page(self):
        url = reverse("edit_thing", kwargs=dict(thing_id=self.thing.id)) + "?add_js_delay=5"
        self.get_literal_url(url)
        self.fill(
            {
                "#id_name": "New name",
                "#id_big": False,
                "#id_clever": True,
                "#id_category_1": Thing.CATEGORY_QUASIGROUP,
                "#id_element_type": Thing.ELEMENT_AIR,
                "#id_count": 5,
                "#id_description": "Soft thing\r\nwith line breaks",
            }
        )
        self.submit("input[name=change]")
        self._assertThingChanged()

    def test_assertTextPresent_auto_wait(self):
        self.get_literal_url(reverse("delayed_appearance") + "?add_js_delay=3")

        # Should automatically wait if we pass within specifying an element not present at first:
        self.assertTextPresent("Hello!", within="#new_stuff")

    def test_assertTextPresent_no_wait(self):
        self.get_literal_url(reverse("test_misc"))
        with pytest.raises(TimeoutException):
            self.assertTextPresent("Hello world", within="p.not-a-real-class")


class TestFuncSeleniumCommonFirefox(FuncSeleniumCommonBase, FirefoxBase):
    pass


class TestFuncSeleniumCommonChrome(FuncSeleniumCommonBase, ChromeBase):
    pass



import conftest
import pytest
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.test import TestCase

from django_functest import FuncSeleniumMixin, FuncWebTestMixin


class WebTestBase(FuncWebTestMixin, TestCase):
    pass


@pytest.mark.selenium
class SeleniumTestBase(FuncSeleniumMixin, StaticLiveServerTestCase):
    driver_name = conftest.BROWSER
    display = conftest.SHOW_BROWSER



import django
from django.test import testcases

if django.VERSION < (2,):

    try:
        from django.core.servers.basehttp import ThreadedWSGIServer
    except ImportError:
        import socketserver

        # See
        #  https://code.djangoproject.com/ticket/20238
        #  https://code.djangoproject.com/ticket/27665
        class ThreadedWSGIServer(socketserver.ThreadingMixIn, testcases.WSGIServer):
            pass

        class MultiThreadedLiveServerThread(testcases.LiveServerThread):
            if django.VERSION < (1, 11, 2):

                def _create_server(self, port):
                    return ThreadedWSGIServer(
                        (self.host, port),
                        testcases.QuietWSGIRequestHandler,
                        allow_reuse_address=False,
                    )

            else:
                # Django 1.11.2 changed the signature
                def _create_server(self):
                    return ThreadedWSGIServer(
                        (self.host, self.port),
                        testcases.QuietWSGIRequestHandler,
                        allow_reuse_address=False,
                    )

    class MultiThreadedLiveServerMixin:
        """
        Mixin for a LiveServerTestCase to make it multi-threaded.
        """

        if django.VERSION < (1, 11):

            @classmethod
            def _create_server_thread(cls, host, possible_ports, connections_override):
                return MultiThreadedLiveServerThread(
                    host,
                    possible_ports,
                    cls.static_handler,
                    connections_override=connections_override,
                )

        else:
            server_thread_class = MultiThreadedLiveServerThread

else:
    # Django 2.0+ has the behavior we want built in
    class MultiThreadedLiveServerMixin:
        pass



from django.db import models

UPLOAD_DIR = "tests/uploads"


# After making changes here, run:
#   ./update_migration.py
class Thing(models.Model):
    ELEMENT_EARTH = "e"
    ELEMENT_WATER = "w"
    ELEMENT_AIR = "a"
    ELEMENT_FIRE = "f"
    ELEMENT_CHOICES = [
        (ELEMENT_EARTH, "Earth"),
        (ELEMENT_WATER, "Water"),
        (ELEMENT_AIR, "Air"),
        (ELEMENT_FIRE, "Fire"),
    ]

    CATEGORY_MAGMA = 0
    CATEGORY_QUASIGROUP = 1
    CATEGORY_MONOID = 2
    CATEGORY_CHOICES = [
        (CATEGORY_MAGMA, "Magma"),
        (CATEGORY_QUASIGROUP, "Quasigroup"),
        (CATEGORY_MONOID, "Monoid"),
    ]

    name = models.CharField(max_length=255)
    big = models.BooleanField(default=False)
    clever = models.BooleanField(default=False)
    element_type = models.CharField(max_length=1, choices=ELEMENT_CHOICES)
    category = models.IntegerField(choices=CATEGORY_CHOICES, default=CATEGORY_MAGMA)
    count = models.IntegerField(default=0)
    description = models.TextField(blank=True)
    notes_file = models.FileField(blank=True, upload_to=UPLOAD_DIR)



import inspect
from unittest import TestCase

import pytest
from django.contrib.auth import get_user_model

from django_functest import AdminLoginMixin, FuncBaseMixin, FuncSeleniumMixin, FuncWebTestMixin, ShortcutLoginMixin

from .base import ChromeBase, FirefoxBase, WebTestBase

LOGGED_OUT_URL = "/admin/login/?next=/admin/"


class ShortcutLoginBase(ShortcutLoginMixin):
    def setUp(self):
        super().setUp()
        User = get_user_model()
        self.user = User.objects.create_superuser("admin", "admin@example.com", "password")

    def test_login_succeeds(self):
        self.shortcut_login(username=self.user.username, password="password")
        self.get_url("admin:index")
        self.assertUrlsEqual("/admin/")

    def test_login_raises_exception_with_wrong_password(self):
        with pytest.raises(ValueError):
            self.shortcut_login(username=self.user.username, password="foo")

    def test_logout_succeeds(self):
        self.shortcut_login(username=self.user.username, password="password")
        self.shortcut_logout()
        self.get_url("admin:index")
        self.assertUrlsEqual(LOGGED_OUT_URL)

    def test_login_by_user(self):
        self.shortcut_login(user=self.user)
        self.get_url("admin:index")
        self.assertUrlsEqual("/admin/")


class TestShortcutLoginWebTest(ShortcutLoginBase, WebTestBase):
    pass


class TestShortcutLoginFirefox(ShortcutLoginBase, FirefoxBase):
    pass


class TestShortcutLoginChrome(ShortcutLoginBase, ChromeBase):
    pass


class AdminLoginBase(AdminLoginMixin):
    def setUp(self):
        super().setUp()
        User = get_user_model()
        self.user = User.objects.create_superuser("admin", "admin@example.com", "password")

    def test_login_succeeds(self):
        self.do_login(username="admin", password="password", shortcut=False)
        self.get_url("admin:index")
        self.assertUrlsEqual("/admin/")

    def test_login_shortcut_succeeds(self):
        self.do_login(username="admin", password="password", shortcut=True)
        self.get_url("admin:index")
        self.assertUrlsEqual("/admin/")

    def test_login_raises_exception_with_wrong_password(self):
        with pytest.raises(ValueError):
            self.do_login(username="admin", password="password_2")

    def test_logout_succeeds(self):
        self.shortcut_login(username="admin", password="password")
        self.do_logout(shortcut=True)
        self.get_url("admin:index")
        self.assertUrlsEqual(LOGGED_OUT_URL)

    def test_logout_shortcut_succeeds(self):
        self.shortcut_login(username="admin", password="password")
        self.do_logout(shortcut=False)
        self.get_url("admin:index")
        self.assertUrlsEqual(LOGGED_OUT_URL)

    def test_shortcut_session_data(self):
        self.do_login(username="admin", password="password", shortcut=False)
        logged_in_session_data = self.get_session_data()
        self.get_url("admin:logout")
        logged_out_session_data = self.get_session_data()
        assert not logged_out_session_data

        # The shortcut method should produce the same session data
        # as the real method.
        self.shortcut_login(username="admin", password="password")
        logged_in_session_data_shortcut = self.get_session_data()
        self.shortcut_logout()
        logged_out_session_data_shortcut = self.get_session_data()
        assert logged_in_session_data == logged_in_session_data_shortcut
        assert logged_out_session_data == logged_out_session_data_shortcut


class TestAdminLoginWebTest(AdminLoginBase, WebTestBase):
    pass


class TestAdminLoginFirefox(AdminLoginBase, FirefoxBase):
    pass


class TestAdminLoginChrome(AdminLoginBase, ChromeBase):
    pass


class TestDocStrings(TestCase):
    def test_doc_strings(self):
        bad_docstrings = []
        for cls in [FuncSeleniumMixin, FuncWebTestMixin]:
            for name, member in inspect.getmembers(cls):
                if name.startswith("__"):
                    continue
                member_doc = getattr(member, "__doc__", None)

                base_member = getattr(FuncBaseMixin, name, None)
                if base_member is not None:
                    base_doc = getattr(base_member, "__doc__", None)
                    if base_doc is not None and member_doc != base_doc:
                        bad_docstrings.append((cls, name))

        if bad_docstrings:
            self.fail(
                "The following methods have incorrect or missing docstrings "
                "compared to FuncBaseMixin: \n" + "\n".join(f"{cls.__name__}.{name}" for cls, name in bad_docstrings)
            )
